#include "blmhistory.h"
#include "ui_blmhistory.h"
#include "image.h"
#include "painter.h"
#include "label.h"
#include "plot/plot.h"
#include <QImage>
#include <QLabel>
#include "db/dbthread.h"
#include <QTWatcher>
#include <QtDebug>
#include <QMessageBox>
#include <QProgressBar>
#include <macros.h>
#include "blmhistory_constants.h"
#include "plot/blmscalelabel.h"
#include "timeintervalhelper.h"
#include <QScrollBar>
#include <QSettings>
#include <QVariant>
#include <QHeaderView>
#include <math.h>

#include "guihelper.h"
#include "windowssaverestore.h"
#include "exportdata.h"

BlmHistory::~BlmHistory()
{
    QSettings s;
    s.setValue("SHOW_DATA_ON_CLICK", ui->cbShowDataOnClick->isChecked());
    s.setValue("RESTORE_START_STOP_DATETIME", ui->cbSaveStartStop->isChecked());
    if(ui->cbSaveStartStop->isChecked())
    {
        s.setValue("START_DATETIME", ui->startDTE->dateTime());
        s.setValue("STOP_DATETIME", ui->stopDTE->dateTime());
    }
    delete ui;
}

BlmHistory::BlmHistory(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    mType = None;
    QString databaseHost = "srv-log-srf.fcs";
    QString databaseName = "hdb";
    QString dbuser = "hdbbrowser", dbpass = "hdbbrowser";
    if(qApp->arguments().count() < 4)
    {
        QMessageBox::information(0, "Usage",
                                 QString("%1: blmmaster1 blmmaster2 blmmaster3 [...]"
                                         "\n[--database=dbname:port] [----type=photodiode|--type=bergoz]\n"
                                         "[--schema-type=hdb++] [--schema=hdbpp] (default schema \"hdb\")\n"
                                         "[--dbuser=username] (default: hdbbrowser) \n"
                                         "[--dbpass=password] (default: hdbbrowser) ").
                                 arg(qApp->arguments().first()));
        exit(EXIT_FAILURE);
    }

    foreach(QString arg, qApp->arguments())
    {
        if(arg.startsWith("--database="))
            databaseHost = arg.replace("--database=", "");
        else if(arg == "--type=photodiode")
            mType = Photodiode;
        else if(arg == "--type=bergoz")
            mType = Bergoz;
        else if(arg.startsWith("--schema="))
            databaseName = arg.remove("--schema=");
        else if(arg.startsWith("--dbuser="))
            dbuser = arg.remove("--dbuser=");
        else if(arg.startsWith("--dbpass="))
            dbpass = arg.remove("--dbpass=");

    }


    ui->setupUi(this);

    WindowsSaveRestore wsr;
    wsr.restore(this);

    bool hdbpp = qApp->arguments().contains("--schema=hdbpp");
    mTotSteps = 0;
    mImageLabel  = new Label(this);
    ui->scrollArea->setWidget(mImageLabel);
    mImageLabel->setPalette("c");

    /* config/tools tab set to index 0 */
    ui->tabWidgetConfig->setCurrentIndex(0);

#if QT_VERSION >= 0x050000
    ui->treeWidgetKeyboardShortcuts->header()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
    ui->treeWidgetKeyboardShortcuts->header()->setResizeMode(QHeaderView::ResizeToContents);
#endif
    /* configure plots */
    ui->plotBlmOnX->setType(Plot::BlmOnX);
    /* time on x */
    ui->plotTimeOnX->setType(Plot::TimeOnX);

    /* QSettings: restore from previous session
     *
     */
    QSettings s;
    ui->cbShowDataOnClick->setChecked(s.value("SHOW_DATA_ON_CLICK", false).toBool());
    if(s.value("RESTORE_START_STOP_DATETIME").toBool())
    {
        ui->startDTE->setDateTime(s.value("START_DATETIME").toDateTime());
        ui->stopDTE->setDateTime(s.value("STOP_DATETIME").toDateTime());
    }
    else /* last 24 hours */
    {
        QDateTime now = QDateTime::currentDateTime();
        ui->stopDTE->setDateTime(now);
        ui->startDTE->setDateTime(now.addDays(-1));
    }

    /*
     * end of QSettings restore
     */

    /* Database thread */
    mDbThread = new DbThread(this, dbuser, dbpass,  databaseHost, databaseName, 3306, hdbpp);
    connect(mDbThread, SIGNAL(finished()), this, SLOT(dbDataAvailable()));
    connect(mDbThread, SIGNAL(progress(int)), this, SLOT(onProgressUpdate(int)), Qt::QueuedConnection);

    QStringList blmMasterDevices = qApp->arguments();
    blmMasterDevices.removeFirst(); /* app name */
    /* remove other command line arguments */
    foreach(QString arg, blmMasterDevices)
        if(arg.startsWith("--"))
            blmMasterDevices.removeAll(arg);

    mBlmNames.resize(blmMasterDevices.size()); /* reserve space for string lists */

    QString cmdName;
    if(mType == Bergoz)
        cmdName = "GetBlmBergozName";
    else if(mType == Photodiode)
        cmdName = "GetBlmPhotodiodesName";
    else
        cmdName = "GetBlmNames";

    foreach(QString blmMasterDev, blmMasterDevices)
    {
        QString src = blmMasterDev + "->" + cmdName;
        QTWatcher *blmNamesWatcher = new QTWatcher(this);
        connect(blmNamesWatcher->qtangoComHandle(), SIGNAL(newData(TVariant)), this, SLOT(blmNamesReady(TVariant)));
        blmNamesWatcher->setSingleShot(true);
        blmNamesWatcher->setSource(src);
    }

    connect(ui->actionExport, SIGNAL(triggered()), this, SLOT(exportData()));
    connect(ui->actionQuit, SIGNAL(triggered()), qApp, SLOT(quit()));
    connect(this, SIGNAL(progress(int)), this, SLOT(onProgressUpdate(int)));
    connect(ui->pbShow, SIGNAL(clicked()), this, SLOT(getData()));
    connect(ui->pbAbort, SIGNAL(clicked()), mDbThread, SLOT(setAbortFlag()));
    connect(ui->tabWidgetConfig, SIGNAL(currentChanged(int)), this, SLOT(configTabChanged(int)));
    ui->pbAbort->setVisible(false);

    connect(mImageLabel, SIGNAL(dataFixedBlm(QString,QVector<double>,QVector<double>)),
            ui->plotTimeOnX, SLOT(updateData(QString,QVector<double>,QVector<double>)));

    connect(mImageLabel, SIGNAL(dataFixedTime(QString,QVector<double>,QVector<double>)),
            ui->plotBlmOnX, SLOT(updateData(QString,QVector<double>,QVector<double>)));

    ui->sbGridAlpha->setValue(mImageLabel->blmGridAlpha());
    ui->cbShowTimeGrid->setDisabled(true); /* enabled only when pixel width is at least 4 in time */
    connect(ui->cbShowTextOnCanvas, SIGNAL(toggled(bool)), mImageLabel, SLOT(setDrawText(bool)));
    connect(ui->cbShowBlmGrid, SIGNAL(toggled(bool)), mImageLabel, SLOT(setShowBlmGrid(bool)));
    connect(ui->sbGridAlpha, SIGNAL(valueChanged(int)), mImageLabel, SLOT(setGridAlpha(int)));
    connect(ui->pbChangeBlmGridColor, SIGNAL(clicked()), mImageLabel, SLOT(toggleGridColor()));
    connect(ui->cbShowTimeGrid, SIGNAL(toggled(bool)), mImageLabel, SLOT(setShowTimeGrid(bool)));
    connect(mImageLabel, SIGNAL(mouseOverLabel(QString)), ui->lBlm, SLOT(setText(QString)));
    connect(mImageLabel, SIGNAL(mouseOverX1(QString)), ui->lTimestamp, SLOT(setText(QString)));
    connect(mImageLabel, SIGNAL(mouseOverX2(QString)), ui->lValue, SLOT(setText(QString)));
    connect(mImageLabel, SIGNAL(maxValueUpdated(double)), ui->lMax, SLOT(setNum(double)));
    connect(mImageLabel, SIGNAL(minValueUpdated(double)), ui->lMin, SLOT(setNum(double)));
    connect(mImageLabel, SIGNAL(maxValueUpdated(double)), this, SLOT(maxValueUpdated(double)));
    connect(mImageLabel, SIGNAL(minValueUpdated(double)), this, SLOT(minValueUpdated(double)));
    connect(mImageLabel, SIGNAL(mouseOverLabel(QString)), ui->lMagnifierBlm, SLOT(setText(QString)));
    connect(mImageLabel, SIGNAL(mouseOverX1(QString)), ui->lMagnifierTimestamp, SLOT(setText(QString)));
    connect(mImageLabel, SIGNAL(mouseOverX2(QString)), ui->lMagnifierValue, SLOT(setText(QString)));
    connect(mImageLabel, SIGNAL(timeGridPossible(bool)), ui->cbShowTimeGrid, SLOT(setEnabled(bool)));
    /* value of empty value */
    connect(ui->pbApplyNullDist, SIGNAL(clicked()), this, SLOT(setEmptyValue()));

    /* dynamic range */
    ui->sbMinLog->setDisabled(true);
    ui->pbApplyMinLog->setDisabled(true);
    foreach(QRadioButton *rb, ui->gbDynamicRange->findChildren<QRadioButton *>())
        connect(rb, SIGNAL(toggled(bool)), this, SLOT(setImageDynamicRange(bool)));
    connect(ui->pbApplyMinLog, SIGNAL(clicked()), this, SLOT(setMinExpLog()));

    /* palettes */
    foreach(QRadioButton *rb, ui->gbPalette->findChildren<QRadioButton *>())
        connect(rb, SIGNAL(toggled(bool)), this, SLOT(changePalette(bool)));

    /* date/time buttons and spinboxes */
    TimeIntervalHelper *timeIntervalHelper =
            new TimeIntervalHelper(ui->tabWidgetConfig->widget(0), ui->centralwidget);
    connect(timeIntervalHelper, SIGNAL(getData()), this, SLOT(getData()));

    connect(ui->cbLive, SIGNAL(toggled(bool)), this, SLOT(setLiveMode(bool)));

    /* magnifier */
    ui->sbMagnifierWidth->setValue(mImageLabel->magnifierAreaWidth());
    ui->sbMagnifierHeight->setValue(mImageLabel->magnifierAreaHeight());
    connect(ui->sbCrossAlpha, SIGNAL(valueChanged(int)), ui->lMagnifier, SLOT(setCrossAlpha(int)));
    connect(this, SIGNAL(magnifierEnabled(bool)), mImageLabel, SLOT(setMagnifierEnabled(bool)));
    connect(ui->sbMagnifierWidth, SIGNAL(valueChanged(int)), mImageLabel, SLOT(setMagnifierAreaWidth(int)));
    connect(ui->sbMagnifierHeight, SIGNAL(valueChanged(int)), mImageLabel, SLOT(setMagnifierAreaHeight(int)));
    connect(mImageLabel, SIGNAL(magnifierHeightChanged(int)), ui->sbMagnifierHeight, SLOT(setValue(int)));
    connect(mImageLabel, SIGNAL(magnifierWidthChanged(int)), ui->sbMagnifierWidth, SLOT(setValue(int)));
    connect(mImageLabel, SIGNAL(crossColorChanged(QColor)), ui->lMagnifier, SLOT(setCrossColor(QColor)));

    /* custom max value */
    ui->dsbMaxValue->setMinimum(BLM_MIN_VALUE);
    ui->dsbMaxValue->setMaximum(BLM_MAX_VALUE);
    ui->dsbMinValue->setMinimum(BLM_MIN_VALUE);
    ui->dsbMinValue->setMaximum(BLM_MAX_VALUE);

    ui->gbCustomMax->setChecked(false);
    connect(ui->gbCustomMax, SIGNAL(toggled(bool)), this, SLOT(enableCustomRange(bool)));
    connect(ui->pbApplyCustomRange, SIGNAL(clicked()), this, SLOT(updateCustomRange()));
    ui->dsbMaxValue->setValue(ui->dsbMaxValue->maximum());
    ui->dsbMinValue->setValue(ui->dsbMinValue->minimum());
    ui->maxValSlider->setMinimum(ui->dsbMaxValue->minimum());
    ui->maxValSlider->setMaximum(ui->dsbMaxValue->maximum());
    ui->maxValSlider->setValue(ui->dsbMaxValue->value());
    connect(ui->maxValSlider, SIGNAL(valueChanged(int)), this, SLOT(setMaxSpinBoxValue(int)));
    connect(ui->dsbMaxValue, SIGNAL(valueChanged(double)), this, SLOT(setMaxValSliderValue(double)));
    connect(ui->pbResetCustomMaxMin, SIGNAL(clicked()), this, SLOT(resetMaxAndMinDoubleSpinboxes()));
}

void BlmHistory::dbDataAvailable()
{
    DbThread *dbt = findChild<DbThread *>();
    if(dbt)
    {
        int maxDataLen;
        QHash<QString, QList<HdbData> > attDataHash = dbt->attDataHash();
        printf("\e[1;32m db data is available: %d elements\e[0m\n", attDataHash.size());

        /* update image */
        QStringList orderedDeviceList;
        foreach(QStringList devs, mBlmNames)
            orderedDeviceList += devs;
        /* remove from the device list the devices not logged into the database */
        foreach(QString missingDevice, dbt->missingDevices())
        {
            orderedDeviceList.removeAll(missingDevice);
            qDebug() << __FUNCTION__ << "removing " << missingDevice << " from ordered device list" <<
                        "orderedDeviceList size is " << orderedDeviceList.size();
        }


        qDebug() << __FUNCTION__ << "mBlmNames" << mBlmNames << " SIZE " << mBlmNames.size();
        qDebug() << __FUNCTION__ << "orderedDeviceList" << orderedDeviceList << " SIZE " << orderedDeviceList.size();

        /* it's time to setup x scale with blm names */
        BlmScaleLabel *blmScaleLabel =
                static_cast<BlmScaleLabel *>(ui->plotBlmOnX->xScaleItem()->scaleLabelInterface());
        if(!blmScaleLabel)
        {
            /* blm names scale label on plot blm on x */
            blmScaleLabel = new BlmScaleLabel();
            ui->plotBlmOnX->xScaleItem()->installScaleLabelInterface(blmScaleLabel);
            ui->plotBlmOnX->xScaleItem()->setTickStepLen(1);
        }
        blmScaleLabel->setBlmDevNames(orderedDeviceList);
        ui->plotBlmOnX->xScaleItem()->setBounds(0, orderedDeviceList.size() - 1);

        maxDataLen = dbt->maxDataLen();
        mImageLabel->setData(attDataHash,
                             orderedDeviceList,
                             maxDataLen,
                             dbt->maximumValue(),
                             dbt->minimumValue(),
                             dbt->minimumPositive(),
                             ui->cbAttName->currentText().toLower());
        mImageLabel->setGeometry(0, 0, IMAGE_W, attDataHash.size() * BLM_PIXWIDTH);

        qDebug() << __FUNCTION__ << "max data len" << dbt->maxDataLen() << orderedDeviceList.size() <<
                    attDataHash.size();
        /* calculate the minimum exponent for the logarithmic dynamic range */
        /* 1. set it on the image label */
        if(ui->rbNatLog->isChecked())
        {
            mImageLabel->setMinExpLog(floor(log(dbt->minimumPositive())) - 2);
            /* 2. update the spin box "min : 1e-N" on the Image tab */
            ui->sbMinLog->setValue(floor(log(dbt->minimumPositive())) - 2);
        }
        else if(ui->rbLog->isChecked())
        {
            mImageLabel->setMinExpLog(floor(log10(dbt->minimumPositive())) - 2);
            /* 2. update the spin box "min : 1e-N" on the Image tab */
            ui->sbMinLog->setValue(floor(log10(dbt->minimumPositive())) - 2);
        }
        mImageLabel->draw();
        GuiHelper guiHelper(this);
        guiHelper.unblock();
        ui->scrollArea->horizontalScrollBar()->setValue(ui->scrollArea->horizontalScrollBar()->maximum());
    }
}

void BlmHistory::blmNamesReady(const TVariant& data)
{
    /* a/b/c->GetBlmNames */
    QStringList blmNames = data.toStringVector().toList();
    QString blmDev = data.tangoPoint().split("->").first();
    /* fill in mBlmNames, vector of QStringList, with the ordered names of the blms */
    for(int i = 1; i < qApp->arguments().count(); i++)
        if(qApp->arguments().at(i).contains(blmDev) && mBlmNames.size() > i - 1)
            mBlmNames[i - 1] = blmNames;

    mBlmNamesHash.insert(data.tangoPoint(), blmNames);
    int sum = 0;
    foreach(QStringList blms, mBlmNamesHash.values())
        sum += blms.size();

    qDebug() << __FUNCTION__ << "total blms" << sum;
}

void BlmHistory::getData()
{
    QDateTime startDt = ui->startDTE->dateTime();
    QDateTime stopDt  = ui->stopDTE->dateTime();
    QString blmAttName = ui->cbAttName->currentText();
    /* get data for each blm */
    mCurrentStep = 0;
    mTotSteps = 0;
    /* busy cursor */
    GuiHelper guiHelper(this);
    guiHelper.blockUI(GuiHelper::GetDataButton);
    QStringList devices;
    foreach(QStringList devs, mBlmNames)
        devices += devs;
    /* put seconds to 0 */
    QTime t1 = startDt.time();
    QTime t2 = stopDt.time();
    t1.setHMS(t1.hour(), t1.minute(), 0, 0);
    t2.setHMS(t2.hour(), t2.minute(), 0, 0);
    startDt.setTime(t1);
    stopDt.setTime(t2);

    mDbThread->getData(devices, blmAttName, startDt, stopDt);
}

void BlmHistory::setLiveMode(bool live)
{
    QTimer *timer = findChild<QTimer *>("liveModeTimer");
    QTimer *ctdt = findChild<QTimer *>("countdownTimer");
    /* to disable configuration widgets during live update */
    GuiHelper guiHelper(this);
    if(live)
    {
        guiHelper.blockUI(GuiHelper::LiveMode);
        if(!timer)
        {
            timer = new QTimer(this);
            timer->setObjectName("liveModeTimer");
            connect(timer, SIGNAL(timeout()), this, SLOT(updateLiveData()));
            ctdt = new QTimer(this);
            ctdt->setObjectName("countdownTimer");
            connect(ctdt, SIGNAL(timeout()), this, SLOT(updateCountdownTimer()));
            ctdt->setInterval(1000);
        }
        ui->lCountdown->setText(QString::number(ui->sbLiveInterval->value()  * 60) + "s");
        timer->setInterval(ui->sbLiveInterval->value() * 60000);
        timer->start();
        ctdt->start();
        /* immediately get data if no data is shown */
        if(mImageLabel->maxDataLen() == 0)
            getData();
    }
    else if(timer)
    {
        timer->stop();
        ctdt->stop();
        ui->lCountdown->setText(ui->sbLiveInterval->value() * 60 + "s");
        DbThread *thread = findChild<DbThread *>();
        if(!thread->isRunning())
            guiHelper.unblock(); /* otherwise the widgets will be enabled when thread finishes */
    }
}

void BlmHistory::resetMaxAndMinDoubleSpinboxes()
{
    ui->dsbMaxValue->setValue(BLM_MAX_VALUE);
    ui->dsbMinValue->setValue(BLM_MIN_VALUE);
}

void BlmHistory::setMaxSpinBoxValue(int val)
{
    ui->dsbMaxValue->setValue(val);
    if(ui->cbApplyMaxWhileSliding->isChecked())
        updateCustomRange();
}

void BlmHistory::setMaxValSliderValue(double val)
{
    ui->maxValSlider->setValue(qRound(val));
}

void BlmHistory::enableCustomRange(bool enab)
{
    mImageLabel->setCustomRangeEnabled(enab);
    if(enab)
        updateCustomRange(); /* calls draw() */
    else
        mImageLabel->draw();
}

void BlmHistory::updateCustomRange()
{
    double max = ui->dsbMaxValue->value();
    double min = ui->dsbMinValue->value();
    mImageLabel->setCustomRange(max, min);
    mImageLabel->draw();
}

void BlmHistory::setEmptyValue()
{
    mImageLabel->emptyValueChanged(ui->sbEmptyVal->value());
}

void BlmHistory::setMinExpLog()
{
    mImageLabel->setMinExpLog(ui->sbMinLog->value());
}

void BlmHistory::setImageDynamicRange(bool dr)
{
    DbThread *dbt = findChild<DbThread* >();
    if(dr) /* take only the checked radio button */
    {
        QString on = sender()->objectName();
        if(on == "rbLinear")
            mImageLabel->setDynamicRange(Painter::Linear);
        if(on == "rbLog")
        {
            mImageLabel->setDynamicRange(Painter::Logarithmic);
            if(dbt)
            {
                mImageLabel->setMinExpLog(floor(log(dbt->minimumPositive())) - 2);
                ui->sbMinLog->setValue(floor(log(dbt->minimumPositive())) - 2);
            }
        }
        else if(on == "rbNatLog")
        {
            mImageLabel->setDynamicRange(Painter::NaturalLog);
            if(dbt)
            {
                mImageLabel->setMinExpLog(floor(log10(dbt->minimumPositive())) - 2);
                ui->sbMinLog->setValue(floor(log10(dbt->minimumPositive())) - 2);
            }
        }
        else if(on == "rbExp")
            mImageLabel->setDynamicRange(Painter::Exponential);
        /* disable/enable elements of the UI according to the checked radio */
        ui->sbEmptyVal->setEnabled(on != "rbLog" && on != "rbNatLog");
        ui->sbMinLog->setEnabled(on == "rbLog" || on == "rbNatLog");
        ui->pbApplyMinLog->setEnabled(on == "rbLog" || on == "rbNatLog");

        if(ui->rbNatLog->isChecked())
        {
            mImageLabel->setMinExpLog(floor(log(dbt->minimumPositive())) - 2);
            /* 2. update the spin box "min : 1e-N" on the Image tab */
            ui->sbMinLog->setValue(floor(log(dbt->minimumPositive())) - 2);
        }
        else if(ui->rbLog->isChecked())
        {
            mImageLabel->setMinExpLog(floor(log10(dbt->minimumPositive())) - 2);
            /* 2. update the spin box "min : 1e-N" on the Image tab */
            ui->sbMinLog->setValue(floor(log10(dbt->minimumPositive())) - 2);
        }
    }
}

void BlmHistory::updateLiveData()
{
    QDateTime newStart = ui->startDTE->dateTime();
    QDateTime newStop = ui->stopDTE->dateTime();
    /* set stop date time equal to current date time and then adjust
     * the start date time according to the current time interval
     * set by the user.
     */
    int deltaSeconds = newStop.secsTo(newStart);
    newStop = QDateTime::currentDateTime();
    newStart = newStop.addSecs(deltaSeconds);
    QTime startT = newStart.time();
    QTime stopT = newStop.time();
    startT = startT.addSecs(ui->sbLiveInterval->value() * 60);
    stopT = stopT.addSecs(ui->sbLiveInterval->value() * 60);
    newStart.setTime(startT);
    newStop.setTime(stopT);
    ui->startDTE->setDateTime(newStart);
    ui->stopDTE->setDateTime(newStop);
    getData();
    ui->lCountdown->setText(QString::number(ui->sbLiveInterval->value() * 60) + "s");
}

void BlmHistory::sourceIsArchived(const QString & , int , int )
{
    mTotSteps ++;
}

void BlmHistory::onProgressUpdate(int step)
{
    QProgressBar *pb = statusBar()->findChild<QProgressBar *>();
    if(!pb && step < 100) /* create and add progress bar */
    {
        pb = new QProgressBar(this);
        pb->setMaximumHeight(20);
        statusBar()->insertWidget(0, pb);
        pb->setMaximum(100);
        pb->setValue(step);
    }
    else if(step == 100 && pb)
    {
        statusBar()->removeWidget(pb);
        delete pb;
    }
    else if(pb)
        pb->setValue(step);
    qApp->processEvents();
}

void BlmHistory::changePalette(bool checked)
{
    if(checked) /* otherwise we call setPalette twice because 2 radio buttons change */
    {
        QString name;
        foreach(QRadioButton *rb, ui->tabWidgetConfig->widget(1)->findChildren<QRadioButton *>())
            if(rb->isChecked())
                name = rb->objectName();
        mImageLabel->setPalette(name.remove("_radioButton"), true); /* true: update image */
    }
}

void BlmHistory::updateCountdownTimer()
{
    QString sval = ui->lCountdown->text();
    sval.remove("s");
    int val = sval.toInt();
    if(val > 0)
        ui->lCountdown->setText(QString::number(--val) + "s");

}

void BlmHistory::exportData() const
{
    ExportData exportData;
    if(!exportData.save(mImageLabel->attDataHash(),
                        mImageLabel->orderedBlmDeviceList(),
                        ui->cbAttName->currentText()))
        QMessageBox::critical(0, "Error saving file", exportData.error());
    else
        exportData.dialogOpenFile();
}

void BlmHistory::configTabChanged(int tabno)
{
    if(tabno == 2)
        connect(mImageLabel, SIGNAL(magnifierRectChanged(const QRect&, const QPixmap*)),
                ui->lMagnifier, SLOT(magnifyArea(const QRect&, const QPixmap*)));
    else
        disconnect(mImageLabel, SIGNAL(magnifierRectChanged(const QRect&, const QPixmap*)),
                   ui->lMagnifier, SLOT(magnifyArea(const QRect&, const QPixmap*)));

    emit magnifierEnabled(tabno == 2);
}

void BlmHistory::closeEvent(QCloseEvent *ce)
{
    WindowsSaveRestore wsr;
    wsr.save(this);
    QMainWindow::closeEvent(ce);
}
