#include "dbdatacache.h"
#include <QtDebug>
#include <macros.h>
#include <QStringList>

DbDataCache::DbDataCache()
{
}

void DbDataCache::invalidate()
{
    mAttDataHash.clear();
    mAttIdHash.clear();
}

/** \brief This method must be called before calculateStartStopDateTime
 *         in order to update the currently start and stop timestamps used by the caller.
 */
void DbDataCache::saveStartStopDateTime(const QDateTime& sdat, const QDateTime stodt)
{
    mStartDt = sdat;
    mStopDt = stodt;
}

int DbDataCache::maxDataLen() const
{
    int max = 0;

    QTime t;
    t.start();
    foreach(QList<HdbData> d, mAttDataHash.values())
    {
        if(d.size() > max)
            max = d.size();
    }
    qDebug() << __FUNCTION__ << "maximum data len calculated in " << t.elapsed() << "ms";
    return max;
}

/** \brief inserts the dataList associated with attName into the data hash,
 *         according to the operation type.
 *
 * @param attName: the name of the attribute to cache
 * @param dataList the list of data associated with attName
 * @param operation the operation type to perform when data is inserted
 *
 * \par Note
 * If the operation is Disjoined and a new insert operation has to be started,
 * the caller must first invalidate the cache
 */
void DbDataCache::insert(const QString& attName,
                         const QList<HdbData> &dataList,
                         OperationType operation)
{
    int dataLen = dataList.size();

    if(!mAttDataHash.contains(attName)) /* insert it */
        mAttDataHash.insert(attName, dataList);
    else if(dataLen > 0)
    {
        /* the att data hash already contains the attName key */
        QList<HdbData> &alreadyExistingList = mAttDataHash[attName];
        int i;
        QDateTime firstExistingTimestamp;
        QDateTime lastExistingTimestamp, timestamp;

        switch(operation)
        {
        case Disjoined:
            /* data hash must have been cleared by the caller before the insertion
             * begins. So we should have fallen into the case
             * if(!mAttDataHash.contains(attName)) above
             */
            perr("DbDataCache: Disjoined operation on a hash already containing \"%s\""
                 ": have you cleared the cache before inserting?",
                 qstoc(attName));
            break;

        case ExtendLeft:

//            qDebug() << __FUNCTION__ << operation << "prepending" << dataList.size() << "to" <<
//                        alreadyExistingList.size() << " adjacent tstamps" <<
//                        alreadyExistingList.first().timestamp << dataList.last().timestamp;

            /* data list is the new data completely preceding the data we already have in
             * the hash: prepend the new data to the old data
             */
            alreadyExistingList = dataList + alreadyExistingList;
            break;

        case ExtendRight:
            if(dataList.size())
            {
//                qDebug() << __FUNCTION__ << operation << "appending" << dataList.size() << "to" <<
//                        alreadyExistingList.size() << " adjacent tstamps" <<
//                        alreadyExistingList.last().timestamp << dataList.first().timestamp;

                alreadyExistingList = alreadyExistingList + dataList;
            }
            break;
        case ExtendLeftRight:
            firstExistingTimestamp = alreadyExistingList.first().timestamp;
            lastExistingTimestamp = alreadyExistingList.last().timestamp;

            /* fill in the left part: insert at the i-th position the i-th value
             * of dataList which timestamp is less than the already existing timestamp
             */
            i = 0;
            timestamp = dataList[i].timestamp;
//            qDebug() << __FUNCTION__ << "joining first ts" << firstExistingTimestamp << "with first new " << timestamp;
            while(timestamp < firstExistingTimestamp && i < dataLen - 1)
            {
                alreadyExistingList.insert(i, dataList[i]);
                i++;
                timestamp = dataList[i].timestamp;
            }

            /* right part.
             * i will point to the first location inside dataList where a timestamp greater than
             * lastExistingTimestamp is expected
             */
            if(i < dataLen)
            {
//                qDebug() << __FUNCTION__ << "joining last ts" << lastExistingTimestamp << "with new " << timestamp;

                while(i < dataLen)
                {
                    if(dataList[i].timestamp > lastExistingTimestamp)
                        alreadyExistingList.append(dataList[i]);
//                    else
//                        qDebug() << __FUNCTION__ << "DbDataCache: data at index %d has timestamp <= " << lastExistingTimestamp;
                    i++;
                }
            }
            else
                printf("DbDataCache: no new data after last existing timestamp\n");



            break;

        case None:
        default:

            break;
        }
    }
}

QList<QDateTime> DbDataCache::calculateStartStopDateTime(const QDateTime &start,
                                                         const QDateTime& stop,
                                                         OperationType &operation)
{
    QList<QDateTime> dtCouples;
    /* first time */
    if(mAttDataHash.isEmpty() ||
            !mStartDt.isValid() ||
            !mStopDt.isValid()
            || stop <= mStartDt
            || start >= mStopDt)
    {
        /* start----------stop    mStartDt++++++mStopDt
         * or
         * mStartDt++++++mStopDt     start------------stop
         */
        qDebug() << __FUNCTION__ << "[0] invalidating cache";
        operation = Disjoined;
        dtCouples << start << stop;
    }
    else if(start < mStartDt && stop > mStopDt)
    {
        /* start-------mStartDt++++++++++++mStopDt------stop */
        qDebug() << __FUNCTION__ << "[1] extending interval";
        operation = ExtendLeftRight;
        dtCouples << start << mStartDt << stop << mStopDt;
    }
    else if(start == mStartDt && stop == mStopDt)
    {
        operation = None;
    }
    else if(start <= mStartDt && stop <= mStopDt)
    {
        /* start------mStartDt+++++++stop++++++mStopDt
         */
        if(start != mStartDt)
            dtCouples << start << mStartDt;
        operation = ExtendLeft;
        /* remove data from stop to mStopDt */
        qDebug() << start << mStartDt << stop << mStopDt;
        foreach(QString blmatt, mAttDataHash.keys())
        {
            QList<HdbData> &dlist = mAttDataHash[blmatt];
            /* remove from the back until the i-th timestamp is greater than stop
             * We keep data with timestamp == stop
             */
            while(dlist.size() > 0 && dlist.last().timestamp > stop)
            {
                qDebug() << __FUNCTION__ << "popping " << dlist.last().timestamp << "cuz > " << stop;
                dlist.pop_back();
            }

            qDebug() << __FUNCTION__ << "[2] removed elements from" << blmatt <<
                        "last timestamp kept" << dlist.last().timestamp << "requested interval" << start << stop;
        }
    }
    else if(start >= mStartDt && stop >= mStopDt)
    {
        /* mStartDt+++++++start++++++mStopDt---------stop */
        if(stop != mStopDt)
            dtCouples << mStopDt << stop;
        operation = ExtendRight;
        /* remove data with timestamp < start */
        foreach(QString blmatt, mAttDataHash.keys())
        {
            QList<HdbData> &dlist = mAttDataHash[blmatt];
            while(dlist.size() > 0 && dlist.at(0).timestamp < start)
                dlist.pop_front();
            if(dlist.size() > 0)
                qDebug() << __FUNCTION__ << "[3] removed elements from" << blmatt <<
                        "first timestamp kept" << dlist.first().timestamp << "requested interval" << start << stop;
        }
    }
    else if(start >= mStartDt && stop <= mStopDt)
    {
        operation = None;
        /* mStartDt++++++start++++++++++++++++++++++stop+++++++++++mStopDt */
        /* no need to make query: will return empty couple but need to remove unused data
         */
        foreach(QString blmatt, mAttDataHash.keys())
        {
            QList<HdbData> &dlist = mAttDataHash[blmatt];
            /* remove from the back until the i-th timestamp is greater than stop
             * We keep data with timestamp == stop
             */
            while(dlist.size() > 0 && dlist.last().timestamp > stop) /* leave timestamp == stop */
                dlist.pop_back();

            qDebug() << __FUNCTION__ << "[4a] removed elements from" << blmatt <<
                        "first timestamp kept" << dlist.first().timestamp << "requested interval" << start << stop;

            while(dlist.size() > 0 && dlist.first().timestamp < start) /* leave timestamp == start */
                dlist.pop_front();

            qDebug() << __FUNCTION__ << "[4b] removed elements from" << blmatt <<
                        "first timestamp kept" << dlist.first().timestamp << "requested interval" << start << stop;
        }
    }
    return dtCouples;
}
