#include "dbthread.h"
#include "dbdatacache.h"
#include "hdbdata.h"
#include "hdbdataaligner.h"
#include <macros.h>
#include <QtDebug>
#include <QtSql/QSqlDatabase>
#include <QtSql/QSqlError>
#include <QtSql/QSqlQuery>
#include <QReadWriteLock>
#include <QMutexLocker>

DbThread::DbThread(QObject *parent,
                   const QString& dbUser,
                   const QString& dbPass,
                   const QString& dbHost,
                   const QString& schema,
                   const int dbPort
                   , bool hdbpp) :
    QThread(parent)
{
    m_dbHost = dbHost;
    m_dbPassword = dbPass;
    m_dbUser = dbUser;
    m_dbPort = dbPort;
    m_hdbPP = hdbpp;
    m_dbSchema = schema;
    mDbDataCache = new DbDataCache();
    mMaximumValue = mMinimumValue = 0.0;
}

void DbThread::getData(const QStringList &devices,
                       const QString&attName,
                       const QDateTime& startDt,
                       const QDateTime& stopDt)
{
    if(attName != mAttName)
        mDbDataCache->invalidate();
    mDeviceList = devices;
    mAttName = attName;
    mStartDt = startDt;
    mStopDt = stopDt;
    this->start();
}

QDateTime DbThread::startDateTime() const
{
    return mStartDt;
}

QDateTime DbThread::stopDateTime() const
{
    return mStopDt;
}

QHash<QString, QList<HdbData> > DbThread::attDataHash()
{
    QReadLocker rlocker(&mRWLock);
    return mDbDataCache->mAttDataHash;
}

void DbThread::setAbortFlag()
{
    QWriteLocker rwLocker(&mAbortRWLock);
    mAbort = true;
}

int DbThread::maxDataLen()
{
    QReadLocker rlocker(&mRWLock);
    return mStartDt.secsTo(mStopDt) / 60 + 1;
}

void DbThread::run()
{
    QString whereTimeQuery;
    QSqlDatabase db = QSqlDatabase::addDatabase("QMYSQL");
    db.setHostName(m_dbHost);
    db.setPassword(m_dbPassword);
    db.setUserName(m_dbUser);
    db.setDatabaseName(m_dbSchema);
    db.setPort(m_dbPort);
    QString attconf_tbl, att_name, att_id, time, value;
    m_hdbPP ? attconf_tbl = "att_conf" : attconf_tbl = "adt";
    m_hdbPP ? att_name = "att_name" : att_name = "full_name";
    m_hdbPP ? att_id = "att_conf_id" : att_id = "ID";
    m_hdbPP ? time = "data_time" : time = "time";
    m_hdbPP ? value = "value_r" : value = "value";

    mAbort = false;

    /* auto reconnect */
    db.setConnectOptions("MYSQL_OPT_RECONNECT=1");

    m_dbError = !db.open();
    if(m_dbError)
    {
        m_lastErrorMessage = QString("HDbProxy: failed to connect to database with host \"%1\", port %2, schema \"%3\""
                                     "user \"%4\" password \"%5\"").
                arg(db.hostName()).arg(db.port()).
                arg(db.databaseName()).arg(db.userName()).
                arg(db.password());

        emit databaseError("DbProxy: db open failed", m_lastErrorMessage);
        perr("database open error: %s", qstoc(m_lastErrorMessage));
    }
    else
    {
        QString attNameLowerCase = mAttName.toLower();
        if(mDeviceList.isEmpty())
        {
            m_lastErrorMessage = QString("HDbProxy: devices list is empty");
            emit databaseError("Error: empty device list", m_lastErrorMessage);
            perr("m_switchDatabase: %s", qstoc(m_lastErrorMessage));
            return;
        }
        QWriteLocker writeLocker(&mRWLock);
        int progressPercent = 0;
        int previousProgressPercent = 0;
        int step = 0;
        int totSteps = 10;
        DbDataCache::OperationType operationType;
        QStringList dateTimeIntervalListAsString;
        QList<QDateTime> dateTimeIntervalList =
                mDbDataCache->calculateStartStopDateTime(mStartDt, mStopDt, operationType);

        /* save current start stop date time */
        mDbDataCache->saveStartStopDateTime(mStartDt, mStopDt);

        /* need to invalidate cache? We have to do it
         */
        if(operationType == DbDataCache::Disjoined)
            mDbDataCache->invalidate();

        foreach(QDateTime dt, dateTimeIntervalList)
            dateTimeIntervalListAsString << dt.toString("yyyy-MM-dd hh:mm");

        if(mDbDataCache->mAttIdHash.isEmpty())
        {
            mMissingDevices.clear();
            pinfo("DbThread: querying database for full_name/ID associations");
            QString part;
            for(int i = 0; i < mDeviceList.size() - 1; i++)
                part += "'%" + mDeviceList.at(i) + "/" + attNameLowerCase + "' or " + att_name + " like ";
            part += "'%" +  mDeviceList.last() + "/" + attNameLowerCase + "'";

            QString queryStr = QString("SELECT %1,%2 FROM %3 where %1 like ")
                    .arg(att_name).arg(att_id).arg(attconf_tbl) + part;

            QSqlQuery q(queryStr);
            printf("\e[1;33mquery for ID: \"%s\" -- SIZE %d\e[0m\n", qstoc(queryStr), q.size());
            qDebug() << __FUNCTION__ << "QUERY " << queryStr;
            if(q.lastError().isValid())
                perr("error in query \"%s\": \"\e[0;31m%s\"\e[0m", qstoc(queryStr), qstoc(q.lastError().text()));
            while(q.next())
            {
                /* remove tango://tango.host.bla.bla and keep only domain/family/member/attribute */
                QString s = q.value(0).toString().section('/', -4, -1).toLower();
                mDbDataCache->mAttIdHash.insert(s, q.value(1).toInt());
                printf("\e[0;35m + inserting in cache attribute \"%s\"\e[0m\n", qstoc(s));
            }

            /* check for missing attributes */
            foreach(QString dev, mDeviceList)
            {
                QString n = dev + "/" + attNameLowerCase;
                if(!mDbDataCache->mAttIdHash.keys().contains(n))
                {
                    perr("DbThread: missing attribute \"%s\" in database", qstoc(n));
                    mMissingDevices << dev;
                }
            }
            if(q.lastError().isValid())
                perr("id query: error %s\n", qstoc(q.lastError().text()));
        }
        else
            pinfo("DbThread: reusing cached full_name/ID couples.");

        emit progress(5);

        totSteps = mDbDataCache->mAttIdHash.size();

        /* compose SQL query now
         * >= and <= only on Disjoined operation.
         * Otherwise, when merging, > and < is enough because the cache will keep the
         * extremes of the interval.
         */
        if(dateTimeIntervalListAsString.size() == 2 && operationType == DbDataCache::Disjoined)
        {
            whereTimeQuery = QString(" WHERE %1 >= '%2' AND %1 <= '%3'").arg(time)
                    .arg(dateTimeIntervalListAsString.first()).arg(dateTimeIntervalListAsString.last());
        }
        else if(dateTimeIntervalListAsString.size() == 2 && operationType == DbDataCache::ExtendLeft)
        {
            whereTimeQuery = QString(" WHERE %1 >= '%2' AND %1 < '%3'").arg(time)
                    .arg(dateTimeIntervalListAsString.first()).arg(dateTimeIntervalListAsString.last());
        }
        else if(dateTimeIntervalListAsString.size() == 2 && operationType == DbDataCache::ExtendRight)
        {
            whereTimeQuery = QString(" WHERE %1 > '%2' AND %1 <= '%3'")
                    .arg(dateTimeIntervalListAsString.first()).arg(dateTimeIntervalListAsString.last());
        }
        else if(dateTimeIntervalListAsString.size() == 4)
        {
            whereTimeQuery = QString(" WHERE (%1 > '%2' AND %1 < '%3') "
                                     " OR (%1 > '%4' AND %1 < '%5')").arg(time)
                    .arg(dateTimeIntervalListAsString.at(0))
                    .arg(dateTimeIntervalListAsString.at(1))
                    .arg(dateTimeIntervalListAsString.at(2))
                    .arg(dateTimeIntervalListAsString.at(3));
        }

        if(!whereTimeQuery.isEmpty() && dateTimeIntervalListAsString.last() > dateTimeIntervalListAsString.first())
        {
            QTime t;
            t.start();
            foreach(QString fullName, mDbDataCache->mAttIdHash.keys())
            {
                int attId = mDbDataCache->mAttIdHash.value(fullName);
                QList<HdbData> hdbDataList;

                mAbortRWLock.lockForRead();
                /* check if we have to stop querying the database */
                if(mAbort)
                {
                    emit progress(100);
                    emit databaseError("DbThread", "Aborted by user");
                    mDbDataCache->invalidate();
                    db.close();
                    mAbortRWLock.unlock();
                    break;
                }
                else
                {
                    QString data_tablename;
                    m_hdbPP ? data_tablename = "att_scalar_devdouble_ro" :
                            data_tablename = QString("att_%1").arg(attId, 5, 10, QChar('0'));
                    QString qu = QString("SELECT %1,%2 FROM %3").arg(time).arg(value).arg(data_tablename);
                    qu += whereTimeQuery;

                    if(m_hdbPP)
                        qu += " AND att_conf_id=" + QString::number(attId);

                    pinfo("executing \"%s\"", qstoc(qu));
                    /* use the 'where' query composed above according to one or two date time intervals
                     * needed
                     */
                    QSqlQuery valuesQuery(qu);
                    if(valuesQuery.lastError().isValid())
                        perr("data query: error %s\n", qstoc(valuesQuery.lastError().text()));

                    while(valuesQuery.next())
                    {
                        QDateTime timestamp = QDateTime::fromString(valuesQuery.value(0).toString(), "yyyy-MM-ddThh:mm:ss");
                        QTime t = timestamp.time();
                        t.setHMS(t.hour(), t.minute(), 0); /* round the value to the minute (discard seconds) */
                        timestamp.setTime(t); /* update date time with a value rounded to the minute */
                        hdbDataList.push_back(HdbData(timestamp, valuesQuery.value(1).toDouble()));
                    }
                    /* data cache insert is optimized according to the operation type to do:
                     * prepend data to the beginning (left), append (right) or extend (left and right)
                     */
                    mDbDataCache->insert(fullName, hdbDataList, operationType);

                    // printf("- inserted %d data for \"%s\" \n", hdbDataList.size(), qstoc(fullName));
                    step++;
                    progressPercent = 100.0 * step / totSteps;
                    /* progressPercent - previousProgressPercent > 9 to avoid progress signal duplicates */
                    if(progressPercent % 10 == 0 && progressPercent - previousProgressPercent > 9)
                    {
                        emit progress(progressPercent);
                        previousProgressPercent = progressPercent;
                    }
                }
                mAbortRWLock.unlock();
            }
            qDebug() << step << "* queries completed in " << t.elapsed() <<"ms, intervals " <<
                        dateTimeIntervalListAsString;
        }
        else
            emit progress(100);
        db.close();

        /* align data into attDataHash according to expected timestamps, one datum each
         * minute.
         * mDbDataCache->mAttDataHash is passed as a reference and so it is modified
         * by the data aligner.
         * HdbDataAligner calculates maximum and minimum values.
         */
        HdbDataAligner dataAligner;
        dataAligner.alignToTimestamps(mDbDataCache->mAttDataHash, mStartDt, mStopDt);
        mMaximumValue = dataAligner.getMaximumValue();
        mMinimumValue = dataAligner.getMinimumValue();
        mMinimumPositive = dataAligner.getMinimumPositive();

        qDebug() << __FUNCTION__ << "max " << mMaximumValue << "min" << mMinimumValue << " min pos "
                 << mMinimumPositive;
    }
}
