#include "label.h"
#include <QPaintEvent>
#include <QPainter>
#include <QtDebug>
#include <macros.h>
#include <QKeyEvent>
#include "image.h"
#include "painter.h"
#include <elettracolors.h>
#include <math.h>
#include "blmhistory_constants.h"
#include "widgets/labeltext.h"

Label::Label(QWidget *parent) :
    QLabel(parent)
{
    mMaxDataLen = 0;
    mMaximumValue = -100;
    mMinimumValue = 1000;
    mMousePos = QPoint();
    /* Creates image data */
    mPainter = new Painter();
    setAlignment(Qt::AlignLeft | Qt::AlignTop);
    setMouseTracking(true);
    mDrawText = false;
    mMagnifierAreaWidth = 100;
    mMagnifierAreaHeight = 100;
    mMagnifierEnabled = false;
    mShowBlmGrid = mShowTimeGrid = false;
    mCrossColors = QList<QColor>() << Qt::green << KDARKBLUE << Qt::white
                                   << Qt::black << Qt::yellow << KLIGHTGRAY;
    mColorIndex = 0;
    mGridColorIndex = mColorIndex + 1;
    mBlmGridAlpha = 160;
    mLogarithmic = mExponential = false;
    mMouseGrabEnabled = true;
    setAlignment(Qt::AlignLeft|Qt::AlignTop);
    QWidget::setPalette(QPalette(Qt::white));
    setText(LABELTEXT); /* labeltext.h */
}

void Label::draw()
{
    mPainter->createData(this);
    if(mPainter->getData() != NULL && mMaxDataLen > 0)
    {
        mText = "";
        QImage image(mPainter->getData(), IMAGE_W, mAttDataHash.size() * BLM_PIXWIDTH, QImage::Format_Indexed8);
        setPalette(mPaletteName);
        image.setColorTable(mColorMap);
        /* Qt documentation: setPixmap clears any previous content
         */
        printf("\e[1;33mbefore setPixmap...\e[0m\n");
        setPixmap(QPixmap::fromImage(image));
        printf("\e[0;33m after setPixmap()\e[0m\n");
        setCursor(QCursor(Qt::BlankCursor));
    }
    else
    {
        mText = "No data in the selected period";
        setPixmap(QPixmap(":images/okguy.jpg"));
        setCursor(QCursor(Qt::ArrowCursor));
    }
}

void Label::setData(const QHash<QString, QList<HdbData> > &attDataHash,
                    const QStringList& orderedDeviceList,
                    int maxDataLen,
                    double maximumValue,
                    double minimumValue,
                    double minimumPositive,
                    const QString &blmAttName)
{
    mAttDataHash = attDataHash;
    mOrderedBlmDeviceList = orderedDeviceList;
    mMaxDataLen = maxDataLen;
    mMaximumValue = maximumValue;
    mMinimumValue = minimumValue;
    mMinimumPositive = minimumPositive;
    mBlmAttName = blmAttName;

    if(!text().isEmpty())
        setText("");

    emit maxValueUpdated(maximumValue);
    emit minValueUpdated(minimumValue);
    emit minPositiveUpdated(minimumPositive);
    emit timeGridPossible(IMAGE_W / mMaxDataLen >= 4);
}

const QHash<QString, QList<HdbData> > & Label::attDataHash() const
{
    qDebug() << __FUNCTION__ << mAttDataHash.size();
    if(mAttDataHash.values().size())
        qDebug() << __FUNCTION__ << mAttDataHash.values().first().size();
    return mAttDataHash;
}

const QStringList & Label::orderedBlmDeviceList() const
{
    return mOrderedBlmDeviceList;
}

void Label::setCustomRangeEnabled(bool en)
{
    mPainter->setCustomRangeEnabled(en);
}

void Label::setCustomRange(double max, double min)
{
     mPainter->setCustomMaxValue(max);
     mPainter->setCustomMinValue(min);
}

void Label::setMinExpLog(int exp)
{
    mPainter->setMinExpLog(exp);
    draw();
}

void Label::invMap(int *blmIndex, int *dataIndex, const QPoint& pos)
{
    if(mMaxDataLen < 1 || pixmap() == NULL)
    {
        *blmIndex = *dataIndex = 0;
    }
    else
    {
        int xi, yi;
        int delta = mMaxDataLen - IMAGE_W;
        int columnStep = IMAGE_W / mMaxDataLen;
        if(delta > 0) /* more data then pixel */
        {
            /* the data completely fits the image */
            xi = pos.x() + delta; /* because the image is top and left - aligned we don't need to calc offsets */
        }
        else
        {
            /* data is filled starting from the right to the left. First image width - datalen * columnStep
             * pixels do not make part of the image.
             */
            xi = pos.x() - (pixmap()->width() - mMaxDataLen * columnStep);
        }

        /* xi is always >= 0 */
        yi = pos.y();
        if(columnStep == 0)
            columnStep = 1;

        *blmIndex = floor(yi / (double)BLM_PIXWIDTH);
        *dataIndex = xi / columnStep;
        //        printf("\e[1;33mdata index xi %d, %d column step %d max datalen %d MIMAGEW %d (%.1f)\e[0m\n",
        //               xi, *dataIndex, columnStep, mMaxDataLen, IMAGE_W, IMAGE_W / (float)mMaxDataLen);
    }
}

void Label::setShowBlmGrid(bool show)
{
    mShowBlmGrid = show;
    update();
}

void Label::emptyValueChanged(int valuePercent)
{
    mPainter->setEmptyValue(valuePercent);
    draw();
}

void Label::enterEvent(QEvent *e)
{
    QLabel::enterEvent(e);
    grabKeyboard();
}

void Label::leaveEvent(QEvent *e)
{
    QLabel::leaveEvent(e);
    mMousePos = QPoint();
    releaseKeyboard();
    update();
}

void  Label::wheelEvent(QWheelEvent *ev)
{
    QLabel::wheelEvent(ev);
    update();
}

void Label::mouseMoveEvent(QMouseEvent *ev)
{
    if(mAttDataHash.size() > 0)
    {
        mMousePos = ev->pos();
        QString val = "-", timestamp = "-";
        int blmindex, dataIndex;
        /* get blm and data indeces */
        invMap(&blmindex, &dataIndex, mMousePos);
        /* get blm name, timestamp and value */
        QString blmAttributeName;
        if(blmindex >= 0 && blmindex < mOrderedBlmDeviceList.size())
            blmAttributeName = mOrderedBlmDeviceList.at(blmindex) + '/' + mBlmAttName;
        else
            blmAttributeName = "No blm";

        /* please before extracting the blm member */
        QList<HdbData> hdbDataList = mAttDataHash.value(blmAttributeName);
        QStringList parts = blmAttributeName.split('/');
        if(parts.size() > 3)
            blmAttributeName = parts.at(2);

        if(dataIndex >= 0 && dataIndex < hdbDataList.size())
        {
            HdbData d = hdbDataList.at(dataIndex);
            val = QString::number(d.value);
            timestamp = d.timestamp.toString("yyyy-MM-dd hh:mm");
        }
        else if(dataIndex < 0 && hdbDataList.size() > 0)
        {
            QDateTime dt = hdbDataList.first().timestamp;
            dt = dt.addSecs(60 * dataIndex);
            timestamp = dt.toString("yyyy-MM-dd hh:mm");
            val = QString::number(nan("NaN"));
        }
        emit mouseOverLabel(blmAttributeName);
        emit mouseOverX1(timestamp);
        emit mouseOverX2(val);
        emit mouseMoved(mMousePos);
        emit magnifierRectChanged(QRect(mMakeMagnifierRect()), pixmap());

        update();
    }
}

void Label::keyPressEvent(QKeyEvent *ev)
{
    int modifiers = ev->modifiers();
    if(mMaxDataLen <= 0)
        QLabel::keyPressEvent(ev);
    else
    {
        int key = ev->key();
        int columnStep = IMAGE_W / mMaxDataLen;
        if(columnStep == 0)
            columnStep = 1;
        QPoint pos = QCursor::pos();
        switch(key)
        {
        case Qt::Key_Up:
            pos.setY(pos.y() - BLM_PIXWIDTH);
            break;
        case Qt::Key_Down:
            pos.setY(pos.y() + BLM_PIXWIDTH);
            break;
        case Qt::Key_Right:
            pos.setX(pos.x() + columnStep);
            break;
        case Qt::Key_Left:
            pos.setX(pos.x() - columnStep);
            break;
        case Qt::Key_C:
            mToggleCrossColor();
            return;
        case Qt::Key_G:
            toggleGridColor();
            return;
        case Qt::Key_H:
            if(modifiers == 0)
                emit magnifierHeightChanged(--mMagnifierAreaHeight);
            else if(modifiers & Qt::ShiftModifier)
                emit magnifierHeightChanged(++mMagnifierAreaHeight);
            return;

        case Qt::Key_W:
            if(modifiers == 0)
                emit magnifierWidthChanged(--mMagnifierAreaWidth);
            else if(modifiers & Qt::ShiftModifier)
                emit magnifierWidthChanged(++mMagnifierAreaWidth);
            return;

        case Qt::Key_M:
            if(mMouseGrabEnabled)
            {
                mMouseGrabEnabled = false;
                setMouseTracking(mMouseGrabEnabled);
                setCursor(Qt::ArrowCursor);
                mText = "Mouse tracking disabled!";
            }
            else
            {
                mMouseGrabEnabled = true;
                setMouseTracking(mMouseGrabEnabled);
                setCursor(Qt::BlankCursor);
                mText.remove("Mouse tracking disabled!");
            }
            update(); /* draw mText */
            return;
        default:
            QLabel::keyPressEvent(ev);
            return;
        }
        mMousePos = pos;
        /* setPos implies mouseMoveEvent, that emits mouseMoved and magnifierRectChanged.
         * No need to emit those signals here though
         */
        QCursor::setPos(pos);
    }
}

void Label::paintEvent(QPaintEvent *e)
{
    QLabel::paintEvent(e);
    if(!mMousePos.isNull() && mAttDataHash.size() > 0)
    {
        const int margin = 3;
        int fontH, fontW;
        double yText, xText;
        QPainter p(this);
        QString val = "-", timestamp = "-";
        QString fullText;
        QColor color;

        if(mMagnifierEnabled)
        {
            color =mCrossColors.at(mColorIndex);
            color.setAlpha(160);
            p.setPen(color);
            p.drawRect(mMakeMagnifierRect());
        }

        color = mCrossColors.at(mColorIndex);
        color.setAlpha(160);
        p.setPen(color);
        /* draw cross on the image, leaving the pixel on the cursor empty.
         * The cursor has been set to Blank when the image is set.
         * In this way, the user should be able to place the mouse exactly over
         * a desired point.
         */
        p.drawLine(0, mMousePos.y(), mMousePos.x() -1, mMousePos.y()); /* 1st hor */
        p.drawLine(mMousePos.x() + 1, mMousePos.y(), width(), mMousePos.y()); /* 2nd hor */

        p.drawLine(mMousePos.x(), 0, mMousePos.x(), mMousePos.y() - 1);
        p.drawLine(mMousePos.x(), mMousePos.y() + 1, mMousePos.x(), height());

        if(mDrawText)
        {
            int blmindex, dataIndex;
            /* get blm and data indeces */
            invMap(&blmindex, &dataIndex, mMousePos);
            /* get blm name, timestamp and value */
            QRect viewportRect = parentWidget()->geometry();
            QString blmAttributeName;
            if(blmindex >= 0 && blmindex < mOrderedBlmDeviceList.size())
                blmAttributeName = mOrderedBlmDeviceList.at(blmindex) + '/' + mBlmAttName;
            else
                blmAttributeName = "No blm";
            /* please before extracting the blm member */
            QList<HdbData> hdbDataList = mAttDataHash.value(blmAttributeName);
            QStringList parts = blmAttributeName.split('/');
            if(parts.size() > 3)
                blmAttributeName = parts.at(2);
            //        qDebug() << __FUNCTION__  << "blmidx" << blmindex << "dataindex" << dataIndex <<
            //                    "data list size" << hdbDataList.size();
            if(dataIndex >= 0 && dataIndex < hdbDataList.size())
            {
                HdbData d = hdbDataList.at(dataIndex);
                val = QString::number(d.value);
                timestamp = d.timestamp.toString("yyyy-MM-dd hh:mm");
            }
            else if(dataIndex < 0 && hdbDataList.size() > 0)
            {
                QDateTime dt = hdbDataList.first().timestamp;
                dt = dt.addSecs(60 * dataIndex);
                timestamp = dt.toString("yyyy-MM-dd hh:mm");
                val = QString::number(nan("NaN"));
            }
            fullText = blmAttributeName + ": " + timestamp + ": " + val;
            QFontMetrics fm(p.font());
            fontH = fm.height() + margin;
            fontW = fm.width(fullText) + margin;
            if(mMousePos.x() + fontW >= viewportRect.width())
                xText = mMousePos.x() - fontW;
            else if(mapToParent(mMousePos).x() - fontW > 0)
                xText = mMousePos.x() - fontW;
            else
                xText = mMousePos.x();

            if(mMousePos.y() - fontH <= 0)
                yText = mMousePos.y() + fontH + margin;
            else
                yText = mMousePos.y() - fontH - margin;

            p.drawText(QPointF(xText, yText), fullText);
        }
    }
    if(!mText.isEmpty())
    {
        QPainter p(this);
        QFont f = p.font();
        f.setBold(true);
        QFontMetrics fm(f);
        p.setPen(KDARKRED);
        p.setFont(f);
        QColor txtBgColor = Qt::white;
        txtBgColor.setAlpha(170);
        QPointF viewPortTopRight = mapToParent(geometry().topRight());
        QRect textRect = QRect(
                    QPoint(width()/2 - fm.width(mText)/2, viewPortTopRight.y() + fm.height()),
                               QSize(fm.width(mText), fm.height()));
        p.fillRect(textRect, txtBgColor);
        p.drawText(textRect,  mText);
    }
    if((mShowBlmGrid || mShowTimeGrid) && pixmap() != NULL)
    {
        const QPixmap *pix = pixmap();
        int step = BLM_PIXWIDTH;
        int columnStep = IMAGE_W / mMaxDataLen;
        QColor gridColor = mCrossColors.at(mGridColorIndex);
        gridColor.setAlpha(mBlmGridAlpha);
        int y = 0, x = pix->width() - columnStep;
        QPainter p(this);
        p.setPen(gridColor);
        while(mShowBlmGrid && y < pix->height())
        {
            p.drawLine(0, y, pix->width(), y);
            y += step;
        }
        while(mShowTimeGrid && x >= columnStep)
        {
            p.drawLine(x, 0, x, pix->height());
            x -= columnStep;
        }
    }
}

void Label::mousePressEvent(QMouseEvent *ev)
{
    if(mAttDataHash.size() > 0)
    {
        double value;
        /* yi individuates the blm */
        int blmindex, dataIndex;
        /* get blm and data indeces */
        invMap(&blmindex, &dataIndex, ev->pos());

        QString blmAttributeName = mOrderedBlmDeviceList.at(blmindex) + '/' + mBlmAttName;
        QList<HdbData> hdbDataList = mAttDataHash.value(blmAttributeName.toLower());


        qDebug() << __FUNCTION__ << ev->pos() << "att name " << blmAttributeName << "blmidx" << blmindex << "dataindex" << dataIndex
                 << "data size" << hdbDataList.size() << "hashSize" << mAttDataHash.size();
        qDebug() << __FUNCTION__  << "data hash keys " << mAttDataHash.keys();
        foreach(QString k, mAttDataHash.keys())
        {
            qDebug() << "data size for " << k << mAttDataHash.value(k).size();
        }


        QVector<double> data, timestamps, valuesInFixedTime, blmIdxs;
        foreach(HdbData d, hdbDataList)
        {
            if(!isnan(d.value))
            {
                data << d.value;
                timestamps << d.timestamp.toTime_t();
            }
        }

        //qDebug() << __FUNCTION__ << blmName << data;
        emit dataFixedBlm(blmAttributeName, timestamps, data);

        /* values for all blms in that time */
        time_t timestamp = 0;
        int idx = 0;
        foreach(QString blm, mOrderedBlmDeviceList)
        {
            blmAttributeName = blm +  '/' + mBlmAttName;
            QList<HdbData> dlist = mAttDataHash.value(blmAttributeName);
            if(dataIndex >= 0 && dlist.size() > dataIndex)
            {
                value = dlist.at(dataIndex).value;
                if(timestamp == 0 && !isnan(value))
                {
                    qDebug() << __FUNCTION__ << "==== converting ====" << dlist.at(dataIndex).timestamp;
                    timestamp = dlist.at(dataIndex).timestamp.toTime_t();
                }
                /* do not want NaNs inside data for plots */
                if(!isnan(value))
                {
                    valuesInFixedTime << dlist.at(dataIndex).value;
                    blmIdxs << idx;
                }
            }
            idx++;
        }
        emit dataFixedTime(QDateTime::fromTime_t(timestamp).toString(), blmIdxs, valuesInFixedTime);

        qDebug() << __FUNCTION__ << QDateTime::fromTime_t(timestamp).toString() << blmIdxs << valuesInFixedTime;
    }
    else
        QLabel::mousePressEvent(ev);
}

QRect Label::mMakeMagnifierRect() const
{
    QPoint center = mMousePos;
    int x0 = center.x() - mMagnifierAreaWidth / 2;
    int y0 = center.y() - mMagnifierAreaHeight / 2;
    return QRect(x0, y0, mMagnifierAreaWidth, mMagnifierAreaHeight);
}

void Label::setDynamicRange(Painter::DynamicRange dr)
{
    mPainter->setDynamicRange(dr);
    draw();
}

void Label::setDrawText(bool draw)
{
    mDrawText = draw;
    update();
}

void Label::setMagnifierEnabled(bool en)
{
    mMagnifierEnabled = en;
    update();
}

void Label::setMagnifierAreaWidth(int width)
{
    mMagnifierAreaWidth = width;
    update();
    emit magnifierRectChanged(QRect(mMakeMagnifierRect()), pixmap());
}

void Label::setMagnifierAreaHeight(int height)
{
    mMagnifierAreaHeight = height;
    update();
    emit magnifierRectChanged(QRect(mMakeMagnifierRect()), pixmap());
}

void Label::mToggleShowCrossText()
{
    mDrawText = !mDrawText;
    update();
}

void Label::toggleGridColor()
{
    mGridColorIndex++;
    mGridColorIndex = mGridColorIndex % 6;
    update();
}

void Label::setGridAlpha(int alpha)
{
    mBlmGridAlpha = alpha;
    update();
}

void Label::mToggleCrossColor()
{
    mColorIndex++;
    mColorIndex = mColorIndex % 6;
    emit crossColorChanged(mCrossColors.at(mColorIndex));
    update();
}

void Label::setShowTimeGrid(bool show)
{
    mShowTimeGrid = show;
    update();
}

void Label::setPalette(const QString &name, bool update)
{
    qDebug() << __FUNCTION__ << "current palette" << mPaletteName << name;
    if(mPaletteName != name)
    {
        mPaletteName = name;
        if(name == "default" || name == "c")
            mMakeDefaultColorMap();
        else if(name == "jet")
            mMakeJetColorMap();
        else if(name == "hot")
            mMakeHotColorMap();
        else if(name == "bw")
            mMakeBWColorMap();
    }
    if(pixmap() && update)
        draw();
}

void Label::mMakeJetColorMap()
{
    int i;
    float deltaB = 1.0/64;
    mColorMap.clear();

    for (i=33; i<65; i++)
    {
        mColorMap.append(QColor::fromRgbF( 0.0, 0.0, (float)i*deltaB ).rgb());
    }
    for (i=1; i<65; i++)
    {
        mColorMap.append(QColor::fromRgbF( 0.0, (float)i*deltaB, 1.0 ).rgb());
    }
    for (i=1; i<65; i++)
    {
        mColorMap.append(QColor::fromRgbF( (float)i*deltaB, 1.0,  1.0-(float)i*deltaB ).rgb());
    }
    for (i=1; i<65; i++)
    {
        mColorMap.append(QColor::fromRgbF( 1.0, 1.0-(float)i*deltaB, 0.0 ).rgb());
    }
    for (i=1; i<33; i++)
    {
        mColorMap.append(QColor::fromRgbF( 1.0-(float)i*deltaB, 0.0, 0.0 ).rgb());
    }
}

void Label::mMakeBWColorMap()
{
    int i;
    mColorMap.clear();
    float r;
    float delta = 1.0/306;
    for (i=0; i < 256; i++)
    {
        r = (float) i * delta;
        mColorMap.append(QColor::fromHsv(0,0,i).rgb());
    }
}

void Label::mMakeHotColorMap()
{
    int i;
    // palette HOT
    mColorMap.clear();
    float deltaA = 1.0/96;
    float deltaB = 1.0/64;
    for (i=1; i<96; i++)
    {
        mColorMap.append(QColor::fromRgbF( (float)i*deltaA, 0.0, 0.0 ).rgb());
    }

    mColorMap.append(QColor::fromRgbF( 1.0, 0.0, 0.0 ).rgb());
    for (i=1; i<96; i++)
    {
        mColorMap.append(QColor::fromRgbF( 1.0, (float)i*deltaA, 0.0 ).rgb());
    }

    mColorMap.append(QColor::fromRgbF( 1.0, 1.0, 0.0 ).rgb());
    for (i=1; i<65; i++)
    {
        mColorMap.append(QColor::fromRgbF( 1.0, 1.0, (float)i*deltaB ).rgb());
    }
}

void Label::mMakeDefaultColorMap()
{
    int i;
    // palette 2
    mColorMap.clear();
    float deltaA = 255.0/31;
    float deltaB = 255.0/62;
    float deltaC = 255.0/37;
    //     int indexTmp = 0;
    for (i=0; i<32; i++) {
        mColorMap.append(QColor( round((float)i*deltaA), 0, round((float)i*deltaA)).rgb());  // (r,g,b)
        //         // qDebug() << indexTmp << "," << i << "," << round((float)i*deltaA) << "," << 0 << "," << round((float)i*deltaA);
        //         indexTmp++;
    }
    for (i=30; i>=0; i--) {
        mColorMap.append(QColor( round(i*deltaA), 0, 255).rgb());  // (r,g,b)
        //         // qDebug() << indexTmp << "," << i << "," << round((float)i*deltaA) << "," << 0 << "," << 255;
        //         indexTmp++;
    }
    for (i=1; i<32; i++) {
        mColorMap.append(QColor( 0, round((float)i*deltaA), 255).rgb());  // (r,g,b)
        //         // qDebug() << indexTmp << "," << i << "," << 0 << "," << round((float)i*deltaA) << "," << 255;
        //         indexTmp++;
    }
    for (i=30; i>=0; i--) {
        mColorMap.append(QColor(0, 255, round((float)i*deltaA)).rgb());  // (r,g,b)
        //         // qDebug() << indexTmp << "," << i << "," << 0 << "," << 255 << "," << round((float)i*deltaA);
        //         indexTmp++;
    }
    for (i=1; i<32; i++) {
        mColorMap.append(QColor( round((float)i*deltaA), 255, 0).rgb());  // (r,g,b)
        //         // qDebug() << indexTmp << "," << i << "," << round((float)i*deltaA) << "," << 255 << "," << 0;
        //         indexTmp++;
    }
    for (i=61; i>=0; i--) {
        mColorMap.append(QColor(255, round((float)i*deltaB), 0).rgb());  // (r,g,b)
        //         // qDebug() << indexTmp << "," << i << "," << 255 << "," << round((float)i*deltaB)<< "," << 0 ;
        //         indexTmp++;
    }
    for (i=1; i<38; i++) {
        mColorMap.append(QColor(255, round((float)i*deltaC), round((float)i*deltaC)).rgb());  // (r,g,b)
        //         // qDebug() << indexTmp << "," << i << "," << 255 << "," << round((float)i*deltaC) << "," << round((float)i*deltaC);
        //         indexTmp++;
    }
    mColorMap.append(QColor(255, 255, 255).rgb());  // (r,g,b)
}
