#include "painter.h"
#include "label.h"
#include <QTime>
#include <QtDebug>
#include <macros.h>
#include <math.h>
#include "blmhistory_constants.h"


Painter::Painter()
{
    mData = NULL;
    mEmptyValue = 10; /* in percentage */
    mDynamicRange = Linear;
    mMinExpLog = -4;
    mCustomMaxValue = -1000;
    mCustomMinValue = 2000;
    mCustomRangeEnabled = false;
}

void Painter::setEmptyValue(int value)
{
    mEmptyValue = value;
}

unsigned char *Painter::getData() const
{
    if(!mData)
        return NULL;
    return mData->data;
}

void Painter::setDynamicRange(DynamicRange dynamicRange)
{
    mDynamicRange = dynamicRange;
}

void Painter::createData(Label *imageLabel)
{
    QTime t;
    t.start();

    /* take some data from imageLabel */
    const QHash<QString, QList<HdbData> >  attDataHash = imageLabel->attDataHash();
    const QStringList orderedBlmDeviceList = imageLabel->orderedBlmDeviceList();
    unsigned char normalizedValue;

    double max, min;
    double minval;
    double range;
    double value;
    double  (*logfunc)(double) = NULL;
    int base;
    int w = IMAGE_W;
    int h = attDataHash.size() * BLM_PIXWIDTH;

    if(!orderedBlmDeviceList.size())
        return;

    if(mCustomRangeEnabled)
    {
        max = mCustomMaxValue;
        minval = mCustomMinValue;
    }
    else
    {
        max = imageLabel->maximumValue();
        minval = imageLabel->minimumValue();
    }


    min = minval - ((max - minval) * mEmptyValue / 100.0);

    qDebug() << __FUNCTION__ << "min" << min << " minval " << minval << "max" << max
             << "mEmptyValue " << mEmptyValue;

    int diffY = 4 - w % 4;
    int nCols = w + diffY % 4;

    if(attDataHash.size() == 0 || imageLabel->maxDataLen() == 0)
        return;

    if(mData)
        delete mData;

    mData = NULL;
    mData = new ImageData(w * nCols);

    double logScaleLessThanZeroMin;

    if(mDynamicRange == Logarithmic)
    {
        logfunc = log10;
        base = 10;
    }
    else if(mDynamicRange == NaturalLog)
    {
        logfunc = log;
        base = M_El;
    }
    if(mDynamicRange == Logarithmic || mDynamicRange == NaturalLog)
    {
        /* preserve the range */
        if(min <= 0)
        {
            min = pow(base, mMinExpLog);
            logScaleLessThanZeroMin = pow(base, mMinExpLog + 1);
            max = max + min; /* shift maximum */
            qDebug() << __FUNCTION__ << min <<"min < 0" << max << min;
        }
        if(max <= 0)
            max = pow(base, mMinExpLog);

        range = logfunc(max) - logfunc(min);
        qDebug() << __FUNCTION__ << max << min << "--> log max" << logfunc(max) <<
                    "log min: " << logfunc(min) << "min exp:" << mMinExpLog << "range" << range;
    }
    else if(mDynamicRange == Exponential)
    {
        range = exp(max) - exp(min);
    }
    else
        range = max - min;

    printf("--> max value is %f min val %f, minbound %f-- took %dms\n2. setting data ncols is %d\n",
           max, min, min, t.elapsed(), nCols);

    t.restart();
    /* fill the image with data. Image is w * h big */
    int j = 0, row, column;
    int i = 0, blmidx, dataidx;

    int columnStep = IMAGE_W / imageLabel->maxDataLen();
    if(columnStep == 0)
        columnStep = 1;

    QString blmattrname;
    int maxidx = 0, minidx = 1e6, idx;

    int printed = 0;
    int minNegPrinted = 0;
    while(i < h)
    {

        blmidx = i / BLM_PIXWIDTH;
        qDebug() << __FUNCTION__ << blmidx << i << h << BLM_PIXWIDTH << "oderedBlmDeviceList size "
                 << orderedBlmDeviceList.size();
        /* pick the name of the blm in the correct order */
        blmattrname = orderedBlmDeviceList.at(blmidx) + "/" + imageLabel->blmAttName();
        /* then pick from the hash the list of hdb data, ordered according to
         * orderedBlmDeviceList
         */
        QList<HdbData> hdbData = attDataHash.value(blmattrname);

        //        printf("blm \"%s\" hdbData size %d\n", qstoc(blmattrname), hdbData.size());

        /* three rows for each blm */
        for(row = i; row < i + BLM_PIXWIDTH; row++) /* BLMs on the row */
        {
            /* make dataidx point to the last location in the data list */
            dataidx = hdbData.size() - 1;

            for(j = w - 1; j - columnStep >= 0; j = j - columnStep)
            {
                /*
                 * column and columnStep:
                 * if there is enough space on the x axis (the data interval is much more
                 * shorter than a day (1440 seconds) ), then we can spread a timestamp over
                 * more than one pixel.
                 * Actually, columnStep pixels
                 */
                column = 0;
                while(column < columnStep)
                {
                    if(dataidx >= 0)
                    {
                        value = hdbData.at(dataidx).value;
                        if(isnan(value))
                        {
                            /// TEST
                            if(printed < 1)
                            {
//                                printf("nan value ---> %f\n", min);
                                printed++;
                            }
                            value = min;
                        }


                        if(value > max)
                            normalizedValue = 255;
                        else if(value < minval)
                        {
//                            printf("value %f < min %f, ", value, min);
                            normalizedValue = nan("NaN");
                        }
                        else if(mDynamicRange == Logarithmic || mDynamicRange == NaturalLog)/* if logarithmic */
                        {
                            if(value <= 0)
                            {
                                value = logScaleLessThanZeroMin;
                                /// TEST
                                if(minNegPrinted < 1)
                                {
//                                    printf("min value ---> %f\n", value);
                                    minNegPrinted++;
                                }
                            }
                            else
                            {
                                /* place value into the logarithmic range */
                                value = logfunc(value) - logfunc(min);
                                normalizedValue = (unsigned char) (qRound(value * 255.0 / range));
                                ///TEST
                                if(printed < 2 && printed == 1)
                                {
//                                    printf("\e[1;32m value of NaN: %g log: %g (normalizeth: %d)\e[0m\n",
//                                           origValue, value, normalizedValue);
                                    printed++;
                                }
                                if(minNegPrinted < 2 && minNegPrinted == 1)
                                {
//                                    printf("\e[1;32m value of min: %g log: %g (normalizeth: %d)\e[0m\n",
//                                           origValue, value, normalizedValue);
                                    minNegPrinted++;
                                }
                            }

                        }
                        else if(mDynamicRange == Exponential)
                        {
                            normalizedValue = (unsigned char) qRound( (exp(value) - exp(min)) * 255.0 / (range));
                        }
                        else
                            normalizedValue = (unsigned char) qRound( (value - min) * 255.0 / (range));
                    }
                    else
                        normalizedValue = nan("NaN");

                    idx = row * nCols  + j - column;
                    mData->data[idx] = normalizedValue;
                    column++;
                    Q_ASSERT(idx > 0);
                    if(idx > maxidx)
                        maxidx = idx;
                    if(idx < minidx)
                        minidx = idx;
                }
                dataidx--;
            }
        }
        i = i + BLM_PIXWIDTH;
    }
//    printf("\n");


    printf("--> data created for image [max idx in data: %d | min %d]. Took %dms\n",
           maxidx, minidx, t.elapsed());
}

